import atexit
import logging
import os
import re
import tarfile
from typing import Any, Dict, List, Optional, Tuple

import docker
import dockerpty
from docker.types import Mount

_logger = logging.getLogger(__name__)
if docker.__version__ < "3.4.0":
    _logger.warning("Unexpected python docker version: %s", docker.__version__)

# TODO detect that user is member of docker group


class DockerClient:
    """Provide additional abstraction and extra methods of the docker client API"""

    client = docker.from_env()

    @classmethod
    def create_volume(
        cls, volume_name: str, extra_labels: Dict[str, str] = None
    ) -> docker.models.volumes.Volume:
        """Return the volume passed in parameter, creating it if it does not exist.
        :param volume_name: name of the volume to create
        :param extra_labels: extra labels to put on the volume (only on
        creation)
        """
        volumes = cls.client.volumes.list(filters={"name": volume_name})
        if volumes:
            _logger.debug("Volume %s already exists", volume_name)
            return volumes[0]
        else:
            _logger.debug("Creating volume %s", volume_name)
            labels = dict(odoo_scripts="")
            if extra_labels:
                labels.update(extra_labels)
            return cls.client.volumes.create(name=volume_name, labels=labels)

    @staticmethod
    def put_file(src, container, dst):
        # copied from https://stackoverflow.com/questions/46390309/how-to-copy-a-file-from-host-to-container-using-docker-py-docker-sdk # noqa: E501
        os.chdir(os.path.dirname(src))
        srcname = os.path.basename(src)
        # TODO put that in a temporary directory
        tar = tarfile.open(src + ".tar", mode="w")
        try:
            tar.add(srcname)
        finally:
            tar.close()

        data = open(src + ".tar", "rb").read()
        container.put_archive(os.path.dirname(dst), data)

    @classmethod
    def exec(
        cls,
        container: docker.models.containers.Container,
        command: List[str],
        user: str,
        interactive: bool = False,
    ) -> int:
        """Execute a command in a container

        :param container: Container to exec on
        :param command: Command to be executed
        :param user: User to execute command as.
        :param interactive: True if the container needs to be interactive
        """
        if interactive:
            dockerpty.exec_command(cls.client.api, container.id, command)
            # it would be nice to return the exec command return code, but
            # there is no way to do that apparently
            result = 0
        else:
            run_result = container.exec_run(command, user=user)
            result = run_result.exit_code
            print(run_result.output.decode())
        return result

    @classmethod
    def run(
        cls,
        repository: str,
        tag: str,
        run_kwargs: Dict[str, Any],
        pull: bool = True,
        target_source_dict: Optional[Dict[str, str]] = None,
    ):
        """Run a container with the provided args"""
        replaces = (
            [(re.compile(key), value) for key, value in target_source_dict.items()]
            if target_source_dict
            else []
        )

        image = f"{repository}:{tag}"

        if pull:
            _logger.debug("Pulling %s", image)
            # TODO find a better way to avoid error when pulling without a connection
            try:
                cls.client.images.pull(repository, tag)
            except docker.errors.APIError:
                _logger.warning(
                    "No connection to server, using existing version, if any"
                )

        container = None

        # XXX When pressing ctrl-c during the container creating, the container
        #  is not removed as there is no return of the value of the Container
        #  object. There is several ways to do that:
        #  - give the container a unique name we know, and use that to clean
        #  - maybe use some kind of time frame to identify the container
        #  - warn the user of a potential container leak by using signals

        def remove():
            """Remove the container"""
            if container:
                _logger.debug("Removing container %s", container.name)
                container.remove()

        def stop_remove():
            """Stop then remove the container."""
            if container:
                if container.status == "running":
                    _logger.debug("Stopping container %s", container.name)
                    container.stop()
                    _logger.debug("Waiting for container %s", container.name)
                    container.wait()
            remove()

        atexit.register(stop_remove)

        # split create and start so in case of SIGTERM, container has fewer chances
        # of being None
        kwargs = run_kwargs.copy()
        kwargs["detach"] = True
        _logger.debug("Create container with image %s", image)
        container = cls.client.containers.create(image, **kwargs)
        _logger.debug("Created container %s", container.name)
        _logger.debug("Starting container %s", container.name)
        container.start()
        atexit.unregister(stop_remove)
        atexit.register(remove)

        for log in container.logs(stream=True, follow=True):
            log = log.decode().rstrip()
            # apply all the replacement of the path inside the image to the real path
            for pattern, value in replaces:
                log = pattern.sub(value, log)
            # print to play nicely with pycharm that will linkify the result
            print(log)

        _logger.debug("Waiting for container %s", container.name)
        result = container.wait()

        if result.get("Error", None):
            _logger.warning(result.decode())
        _logger.debug("Removing container %s", container.name)
        container.remove(v=True)
        atexit.unregister(remove)
        return result["StatusCode"]


def modules_mount(
    modules, project_path, target_path: str
) -> Tuple[List[Mount], Dict[str, str]]:
    """Return a list of Mount for the given modules"""
    mount_list: List[Mount] = []
    mount_dict: Dict[str, str] = {}

    for module in modules:
        full_target_path = os.path.join(target_path, os.path.basename(module))
        full_project_path = os.path.realpath(os.path.join(project_path, module))
        mount_list.append(Mount(full_target_path, full_project_path, "bind"))
        mount_dict[full_target_path] = full_project_path

    return mount_list, mount_dict
